/*
 * This file is part of the FDNPKG project.
 * Copyright (C) Mateusz Viste 2013
 * 
 * Provides a simplistic HTTP / Gopher downloader.
 */

#include <conio.h>  /* ScreenSetCursor() */
#include <stdio.h>
#include <stdlib.h>  /* malloc() */
#include <string.h>   /* strlen() */
#include <time.h>
#include "net.h"
#include "parseurl.h"
#include "http.h"  /* adding self for control */

#define buffer_max 4096


/* a function to fetch the next line from the socket  */
static int sock_getline(struct net_tcpsocket *socket, char *buf, int len) {
  int  i = 0, net_result;
  char ch;
  /* int status = 0;
  int *statusptr = &status; */
  while (i < len) {
    net_result = net_recv(socket, &ch, 1);
    if (net_result == 0) continue;  /* if got nothing, loop again */
    if (net_result < 0) break; /* conn error */
    if (ch == '\n') break;
    if (ch != '\r') {
      *buf++ = ch;
      i++;
    }
  }
  *buf = '\0';
  return(i);
}


/* returns 0 on success, non-zero otherwise
   Warning: Requires net_init() to be run before! */
int http_get(char *orgurl, char *outputfile, int progressfeedbackflag) {
  char *host = NULL;
  char *path = NULL;
  char *buffer = NULL;
  int port, result = -1, proto;
  unsigned long ipaddr;
  int x;
  long fdlen = 0, byteread = 0;
  struct net_tcpsocket *socket;
  FILE *fd;
  time_t curtime, lastrefreshtime;

  /* valide input data */
  if ((orgurl == NULL) || (outputfile == NULL)) return(-1);

  lastrefreshtime = time(NULL) + 1;

  for (;;) { /* not a real loop, just a container allowing to break out easily */
    x = strlen(orgurl) + 1;
    host = malloc(x);
    if (host == NULL) return(-2);
    path = malloc(x);
    if (path == NULL) break;
    buffer = malloc(buffer_max);
    if (buffer == NULL) break;

    proto = parseurl(orgurl, host, &port, path);
    if ((proto != PARSEURL_PROTO_HTTP) && (proto != PARSEURL_PROTO_GOPHER)) break;

    ipaddr = net_dnsresolve(host);
    if (ipaddr == 0) break;

    /* printf("%s -> %lu.%lu.%lu.%lu (%d)\n", host, (ipaddr >> 24) & 0xFF, (ipaddr >> 16) & 0xFF, (ipaddr >> 8) & 0xFF, ipaddr & 0xFF, port); */

    socket = net_connect(ipaddr, port);

    if (socket == NULL) break;

    if (proto == PARSEURL_PROTO_HTTP) {
        sprintf(buffer, "GET /%s HTTP/1.0\r\nHOST: %s\r\nUSER-AGENT: FDNPKG\r\n\r\n", path, host);
      } else {
        sprintf(buffer, "%s\r\n", path);
    }

    if (net_send(socket, buffer, strlen(buffer)) != (int)strlen(buffer)) {
      net_abort(socket);
      break;
    }

    fd = fopen(outputfile, "wb");
    if (fd == NULL) { /* this should not fail */
      fclose(fd);
      net_abort(socket);
      break;
    }

    /* eat all headers out if using http */
    if (proto == PARSEURL_PROTO_HTTP) {
      while (sock_getline(socket, buffer, buffer_max) > 0);
    }

    for (;;) {
      byteread = net_recv(socket, buffer, buffer_max);
      curtime = time(NULL);
      if (((curtime - lastrefreshtime > 1) || (byteread < 0)) && (progressfeedbackflag != 0)) {
        lastrefreshtime = curtime;
        printf("Downloading... %ld bytes", fdlen);
        puts("");
      }
      if (byteread < 0) break;
      fdlen += byteread;
      fwrite(buffer, byteread, 1, fd);
    }
    if (fdlen > 0) { /* close the connection nicely if we got at least 1 byte of data */
        net_close(socket);
      } else { /* otherwise abort asap, the remote server is either non existant or in really bad shape anyway */
        net_abort(socket);
    }
    fclose(fd);
    result = fdlen;
    break;
  }
  if (host != NULL) free(host);
  if (path != NULL) free(path);
  if (buffer != NULL) free(buffer);
  return(result);
}
